import os
from pathlib import Path

from click.testing import CliRunner
from xsdata.cli import cli
from xsdata.models.config import GeneratorConfig
from xsdata.utils.testing import ClassFactory
from xsdata.utils.testing import FactoryTestCase

from xsdata_attrs.generator import AttrsGenerator


class AttrsGeneratorTests(FactoryTestCase):
    def setUp(self):
        super().setUp()
        config = GeneratorConfig()
        self.generator = AttrsGenerator(config)

    def test_render(self):
        classes = [
            ClassFactory.elements(2, package="foo"),
            ClassFactory.elements(3, package="foo"),
        ]

        classes[0].attrs[0].restrictions.max_occurs = 3

        iterator = self.generator.render(classes)

        actual = [(out.path, out.title, out.source) for out in iterator]
        expected = (
            "from typing import List, Optional\n"
            "\n"
            "import attr\n"
            "\n"
            '__NAMESPACE__ = "xsdata"\n'
            "\n"
            "\n"
            "@attr.s\n"
            "class ClassB:\n"
            "    class Meta:\n"
            '        name = "class_B"\n'
            "\n"
            "    attr_b: List[str] = attr.ib(\n"
            "        factory=list,\n"
            "        metadata={\n"
            '            "name": "attr_B",\n'
            '            "type": "Element",\n'
            '            "max_occurs": 3,\n'
            "        },\n"
            "    )\n"
            "    attr_c: Optional[str] = attr.ib(\n"
            "        default=None,\n"
            "        metadata={\n"
            '            "name": "attr_C",\n'
            '            "type": "Element",\n'
            "        },\n"
            "    )\n"
            "\n"
            "\n"
            "@attr.s\n"
            "class ClassC:\n"
            "    class Meta:\n"
            '        name = "class_C"\n'
            "\n"
            "    attr_d: Optional[str] = attr.ib(\n"
            "        default=None,\n"
            "        metadata={\n"
            '            "name": "attr_D",\n'
            '            "type": "Element",\n'
            "        },\n"
            "    )\n"
            "    attr_e: Optional[str] = attr.ib(\n"
            "        default=None,\n"
            "        metadata={\n"
            '            "name": "attr_E",\n'
            '            "type": "Element",\n'
            "        },\n"
            "    )\n"
            "    attr_f: Optional[str] = attr.ib(\n"
            "        default=None,\n"
            "        metadata={\n"
            '            "name": "attr_F",\n'
            '            "type": "Element",\n'
            "        },\n"
            "    )\n"
        )

        self.assertEqual(2, len(actual))
        self.assertEqual(3, len(actual[1]))

        self.assertEqual("foo.tests", actual[1][1])
        self.assertEqual(expected, actual[1][2])

    def test_with_unsafe_hash(self):
        config = GeneratorConfig()
        config.output.format.unsafe_hash = True
        self.generator = AttrsGenerator(config)

        iterator = self.generator.render(ClassFactory.list(1))
        actual = [(out.path, out.title, out.source) for out in iterator]
        self.assertTrue("@attr.s(hash=True)" in actual[1][2])

    def test_complete(self):
        runner = CliRunner()
        schema = Path(__file__).parent.joinpath("fixtures/schemas/po.xsd")
        os.chdir(Path(__file__).parent.parent)

        result = runner.invoke(
            cli,
            [
                str(schema),
                "--package",
                "tests.fixtures.po.models",
                "--structure-style=single-package",
                "--output",
                "attrs",
                "--config",
                "tests/fixtures/attrs.conf.xml",
            ],
        )

        self.assertIsNone(result.exception)



from datetime import datetime
from typing import List

import attr


@attr.s(auto_attribs=True)
class TypeA:
    one: str
    two: float


@attr.s(auto_attribs=True)
class TypeB(TypeA):
    one: str
    three: bool = attr.ib(default=True)


@attr.s(auto_attribs=True)
class TypeC(TypeB):
    four: List[datetime] = attr.ib(factory=list, metadata={"format": "%d %B %Y %H:%M"})
    any: object = attr.ib(default=None, metadata={"type": "Wildcard"})



from decimal import Decimal
from typing import List, Optional

import attr
from xsdata.models.datatype import XmlDate

__NAMESPACE__ = "foo"


@attr.s(slots=True, kw_only=True)
class Usaddress:
    class Meta:
        name = "USAddress"

    name: str = attr.ib(
        metadata={
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    street: str = attr.ib(
        metadata={
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    city: str = attr.ib(
        metadata={
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    state: str = attr.ib(
        metadata={
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    zip: Decimal = attr.ib(
        metadata={
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    country: str = attr.ib(
        init=False,
        default="US",
        metadata={
            "type": "Attribute",
        },
    )


@attr.s(slots=True, kw_only=True)
class Comment:
    class Meta:
        name = "comment"
        namespace = "foo"

    value: str = attr.ib(
        default="",
        metadata={
            "required": True,
        },
    )


@attr.s(slots=True, kw_only=True)
class Items:
    item: List["Items.Item"] = attr.ib(
        factory=list,
        metadata={
            "type": "Element",
            "namespace": "foo",
        },
    )

    @attr.s(slots=True, kw_only=True)
    class Item:
        product_name: str = attr.ib(
            metadata={
                "name": "productName",
                "type": "Element",
                "namespace": "foo",
                "required": True,
            }
        )
        quantity: int = attr.ib(
            metadata={
                "type": "Element",
                "namespace": "foo",
                "required": True,
                "max_exclusive": 100,
            }
        )
        usprice: Decimal = attr.ib(
            metadata={
                "name": "USPrice",
                "type": "Element",
                "namespace": "foo",
                "required": True,
            }
        )
        comment: Optional[Comment] = attr.ib(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "foo",
            },
        )
        ship_date: Optional[XmlDate] = attr.ib(
            default=None,
            metadata={
                "name": "shipDate",
                "type": "Element",
                "namespace": "foo",
            },
        )
        part_num: str = attr.ib(
            metadata={
                "name": "partNum",
                "type": "Attribute",
                "required": True,
                "pattern": r"\d{3}-[A-Z]{2}",
            }
        )


@attr.s(slots=True, kw_only=True)
class PurchaseOrderType:
    ship_to: Usaddress = attr.ib(
        metadata={
            "name": "shipTo",
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    bill_to: Usaddress = attr.ib(
        metadata={
            "name": "billTo",
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    comment: Optional[Comment] = attr.ib(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "foo",
        },
    )
    items: Items = attr.ib(
        metadata={
            "type": "Element",
            "namespace": "foo",
            "required": True,
        }
    )
    order_date: Optional[XmlDate] = attr.ib(
        default=None,
        metadata={
            "name": "orderDate",
            "type": "Attribute",
        },
    )


@attr.s(slots=True, kw_only=True)
class PurchaseOrder(PurchaseOrderType):
    class Meta:
        name = "purchaseOrder"
        namespace = "foo"



from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar

import attr
from xsdata.exceptions import XmlContextError
from xsdata.formats.dataclass.compat import ClassType
from xsdata.formats.dataclass.models.elements import XmlType

T = TypeVar("T", bound=object)


@attr.s
class AnyElement:
    """Generic model to bind xml document data to wildcard fields.

    Args:
        qname: The element's qualified name
        text: The element's text content
        tail: The element's tail content
        children: The element's list of child elements.
        attributes: The element's key-value attribute mappings.
    """

    qname: Optional[str] = attr.ib(default=None)
    text: Optional[str] = attr.ib(default=None)
    tail: Optional[str] = attr.ib(default=None)
    children: List[object] = attr.ib(factory=list, metadata={"type": XmlType.WILDCARD})
    attributes: Dict[str, str] = attr.ib(
        factory=dict, metadata={"type": XmlType.ATTRIBUTES}
    )


@attr.s
class DerivedElement(Generic[T]):
    """Generic model wrapper for type substituted elements.

    Example: eg. <b xsi:type="a">...</b>

    Args:
        qname: The element's qualified name
        value: The wrapped value
        type: The real xsi:type
    """

    qname: str = attr.ib()
    value: T = attr.ib()
    type: Optional[str] = attr.ib(default=None)


class Attrs(ClassType):
    """The attrs class type."""

    @property
    def any_element(self) -> Type:
        """Return the AnyElement used to bind wildcard element nodes."""
        return AnyElement

    @property
    def derived_element(self) -> Type:
        """Return the DerivedElement used to bind ambiguous element nodes."""
        return DerivedElement

    def is_model(self, obj: Any) -> bool:
        """Return whether the given value is binding model."""
        return attr.has(obj if isinstance(obj, type) else type(obj))

    def verify_model(self, obj: Any):
        """Verify the given value is a binding model.

        Args:
            obj: The input model instance

        Raises:
            XmlContextError: if not supported
        """
        if not self.is_model(obj):
            raise XmlContextError(f"Type '{obj}' is not an attrs model.")

    def get_fields(self, obj: Any) -> Tuple[Any, ...]:
        """Return the models fields in the correct mro ordering."""
        if not isinstance(obj, type):
            return self.get_fields(type(obj))

        # Emulate dataclasses fields ordering
        fields = {}
        for b in obj.__mro__[-1:0:-1]:
            if self.is_model(b):
                for f in self.get_fields(b):
                    fields[f.name] = f

        for f in attr.fields(obj):
            fields[f.name] = f

        return tuple(fields.values())

    def default_value(
        self, field: attr.Attribute, default: Optional[Any] = None
    ) -> Any:
        """Return the default value or factory of the given model field."""
        res = field.default
        if res is attr.NOTHING:
            return default

        if isinstance(res, attr.Factory):  # type: ignore
            return res.factory  # type: ignore

        return res

    def default_choice_value(self, choice: Dict) -> Any:
        """Return the default value or factory of the given model field choice."""
        factory = choice.get("factory")
        if callable(factory):
            return factory

        return choice.get("default")



from dataclasses import dataclass
from dataclasses import field
from typing import Callable, Optional

from xsdata.formats.dataclass import context
from xsdata.formats.dataclass import parsers
from xsdata.formats.dataclass import serializers
from xsdata.utils.constants import return_input


class XmlContext(context.XmlContext):
    """The models context class.

    The context is responsible to provide binding metadata
    for models and their fields.

    Args:
        element_name_generator: Default element name generator
        attribute_name_generator: Default attribute name generator
        models_package: Restrict auto locate to a specific package

    Attributes:
        cache: Internal cache for binding metadata instances
        xsi_cache: Internal cache for xsi types to class locations
        sys_modules: The number of loaded sys modules
    """

    def __init__(
        self,
        element_name_generator: Callable = return_input,
        attribute_name_generator: Callable = return_input,
        models_package: Optional[str] = None,
    ):
        super().__init__(
            element_name_generator, attribute_name_generator, "attrs", models_package
        )


@dataclass
class XmlParser(parsers.XmlParser):
    """Default Xml parser for attrs.

    Args:
        config: The parser config instance
        context: The xml context instance
        handler: The xml handler class

    Attributes:
        ns_map: The parsed namespace prefix-URI map
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class XmlSerializer(serializers.XmlSerializer):
    """Xml serializer for attrs.

    Args:
        config: The serializer config instance
        context: The models context instance
        writer: The xml writer class
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class JsonParser(parsers.JsonParser):
    """Json parser for attrs.

    Args:
        config: Parser configuration
        context: The models context instance
        load_factory: Json loader factory
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class DictDecoder(parsers.DictDecoder):
    """Bind a dictionary or a list of dictionaries to attrs.

    Args:
        config: Parser configuration
        context: The models context instance
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class DictEncoder(serializers.DictEncoder):
    """Bind a dictionary or a list of dictionaries to attrs.

    Args:
        config: Parser configuration
        context: The models context instance
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class JsonSerializer(serializers.JsonSerializer):
    """Json serializer for attrs.

    Args:
        config: The serializer config instance
        context: The models context instance
        dict_factory: Dictionary factory
        dump_factory: Json dump factory e.g. json.dump
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class UserXmlParser(parsers.UserXmlParser):
    """Xml parser for attrs with hooks to events.

    The event hooks allow custom parsers to inject custom
    logic between the start/end element events.

    Args:
        config: The parser config instance
        context: The xml context instance
        handler: The xml handler class

    Attributes:
        ns_map: The parsed namespace prefix-URI map
        hooks_cache: The hooks cache is used to avoid
            inspecting the class for custom methods
            on duplicate events.
    """

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class TreeParser(parsers.TreeParser):
    """Bind xml nodes to a tree of AnyElement objects."""

    context: XmlContext = field(default_factory=XmlContext)


@dataclass
class PycodeSerializer(serializers.PycodeSerializer):
    """Pycode serializer for data class instances.

    Generate python pretty representation code from a model instance.

    Args:
        context: The models context instance
    """

    context: XmlContext = field(default_factory=XmlContext)



from typing import Dict
from typing import List
from typing import Optional

from xsdata.codegen.models import Attr
from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.dataclass.generator import DataclassGenerator
from xsdata.models.config import GeneratorConfig
from xsdata.models.config import OutputFormat


class AttrsGenerator(DataclassGenerator):
    """Python attrs classes code generator.

    Args:
        config: The generator config instance

    Attributes:
        env: The jinja2 environment instance
        filters: The template filters instance
    """

    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        """Initialize the filters instance by the generator configuration."""
        return AttrsFilters(config)


class AttrsFilters(Filters):
    FACTORY_KEY = "factory"
    DEFAULT_KEY = "default"

    @classmethod
    def build_class_annotation(cls, format: OutputFormat) -> str:
        """Build the class annotations."""
        result = super().build_class_annotation(format)
        result = result.replace("unsafe_hash=", "hash=")
        result = result.replace("@dataclass", "@attr.s")
        return result

    def field_definition(
        self,
        attr: Attr,
        ns_map: Dict,
        parent_namespace: Optional[str],
        parents: List[str],
    ) -> str:
        """Return the field definition with any extra metadata."""
        result = super().field_definition(attr, ns_map, parent_namespace, parents)
        return result.replace("field(", "attr.ib(")

    @classmethod
    def build_import_patterns(cls) -> Dict[str, Dict]:
        """Build import search patterns."""
        patterns = {"attr": {"__module__": [" attr.ib", "@attr.s"]}}
        patterns.update(super().build_import_patterns())

        return patterns



from datetime import datetime
from unittest import TestCase

from xsdata.formats.dataclass.serializers.config import SerializerConfig

from tests.fixtures.common import TypeC
from xsdata_attrs.bindings import JsonParser
from xsdata_attrs.bindings import JsonSerializer
from xsdata_attrs.bindings import XmlParser
from xsdata_attrs.bindings import XmlSerializer
from xsdata_attrs.compat import AnyElement
from xsdata_attrs.compat import DerivedElement


class BindingsTests(TestCase):
    def setUp(self):
        self.obj = TypeC(
            one="first",
            two=1.1,
            four=[
                datetime(2002, 1, 1, 12, 1),
                datetime(2003, 2, 5, 13, 5),
            ],
            any=AnyElement(
                children=[
                    AnyElement(qname="foo", text="bar"),
                    DerivedElement(qname="bar", value="1"),
                    DerivedElement(qname="bar", value=2),
                ]
            ),
        )

    def test_xml_bindings(self):
        serializer = XmlSerializer()
        serializer.config.indent = "  "
        serializer.config.xml_declaration = False
        parser = XmlParser()
        ns_map = {
            "xs": "http://www.w3.org/2001/XMLSchema",
            "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        }

        expected = (
            '<TypeC xmlns:xs="http://www.w3.org/2001/XMLSchema" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\n'
            "  <one>first</one>\n"
            "  <two>1.1</two>\n"
            "  <three>true</three>\n"
            "  <four>01 January 2002 12:01</four>\n"
            "  <four>05 February 2003 13:05</four>\n"
            "  <foo>bar</foo>\n"
            '  <bar xsi:type="xs:string">1</bar>\n'
            '  <bar xsi:type="xs:short">2</bar>\n'
            "</TypeC>\n"
        )
        self.assertEqual(expected, serializer.render(self.obj, ns_map))
        self.assertEqual(self.obj, parser.from_string(expected))

    def test_serialize_json(self):
        serializer = JsonSerializer(config=SerializerConfig(indent="  "))
        parser = JsonParser()

        expected = (
            "{\n"
            '  "one": "first",\n'
            '  "two": 1.1,\n'
            '  "three": true,\n'
            '  "four": [\n'
            '    "01 January 2002 12:01",\n'
            '    "05 February 2003 13:05"\n'
            "  ],\n"
            '  "any": {\n'
            '    "qname": null,\n'
            '    "text": null,\n'
            '    "tail": null,\n'
            '    "children": [\n'
            "      {\n"
            '        "qname": "foo",\n'
            '        "text": "bar",\n'
            '        "tail": null,\n'
            '        "children": [],\n'
            '        "attributes": {}\n'
            "      },\n"
            "      {\n"
            '        "qname": "bar",\n'
            '        "value": "1",\n'
            '        "type": null\n'
            "      },\n"
            "      {\n"
            '        "qname": "bar",\n'
            '        "value": 2,\n'
            '        "type": null\n'
            "      }\n"
            "    ],\n"
            '    "attributes": {}\n'
            "  }\n"
            "}"
        )
        self.assertEqual(expected, serializer.render(self.obj))
        self.assertEqual(self.obj, parser.from_string(expected, TypeC))



from datetime import datetime
from operator import attrgetter
from unittest import TestCase

from xsdata.exceptions import XmlContextError
from xsdata.formats.dataclass.compat import class_types

from tests.fixtures.common import TypeC
from xsdata_attrs.bindings import XmlContext
from xsdata_attrs.compat import AnyElement
from xsdata_attrs.compat import Attrs
from xsdata_attrs.compat import DerivedElement


class AttrsTests(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.class_type = class_types.get_type("attrs")

    def test_class_type(self):
        self.assertIsInstance(self.class_type, Attrs)

    def test_property_any_element(self):
        self.assertIs(self.class_type.any_element, AnyElement)

    def test_property_derived_element(self):
        self.assertIs(self.class_type.derived_element, DerivedElement)

    def test_is_model(self):
        obj = AnyElement()
        self.assertTrue(self.class_type.is_model(obj))
        self.assertTrue(self.class_type.is_model(AnyElement))
        self.assertFalse(self.class_type.is_model(XmlContext))

    def test_verify_model(self):
        obj = AnyElement()
        self.class_type.verify_model(obj)

        with self.assertRaises(XmlContextError) as cm:
            self.class_type.verify_model(int)

        self.assertEqual(
            "Type '<class 'int'>' is not an attrs model.", str(cm.exception)
        )

    def test_get_fields(self):
        obj = TypeC(one="a", two=1.1, three=False, four=datetime.now())
        expected = ("one", "two", "three", "four", "any")
        get_name = attrgetter("name")

        from_obj = self.class_type.get_fields(obj)
        self.assertEqual(expected, tuple(map(get_name, from_obj)))

        from_type = self.class_type.get_fields(TypeC)
        self.assertEqual(expected, tuple(map(get_name, from_type)))

    def test_default_value(self):
        fields = self.class_type.get_fields(TypeC)

        self.assertIsNone(self.class_type.default_value(fields[0]))
        self.assertIsNone(self.class_type.default_value(fields[1]))
        self.assertTrue(self.class_type.default_value(fields[2]))
        self.assertEqual(list, self.class_type.default_value(fields[3]))

    def test_default_choice_value(self):
        choice = {}
        self.assertIsNone(self.class_type.default_choice_value(choice))

        choice["factory"] = 1
        self.assertIsNone(self.class_type.default_choice_value(choice))

        choice["default"] = 1
        self.assertEqual(1, self.class_type.default_choice_value(choice))

        choice["factory"] = lambda: 2

        default = self.class_type.default_choice_value(choice)
        self.assertTrue(callable(default))
        self.assertEqual(2, default())
